import urllib.request as request
from urllib.parse import urlparse
import os
import uuid

from notion2md.console.formatter import *
from notion2md.config import Config

from clikit.api.io import IO

# Since external file block cannot guarantees
# whether the file is downloadable or not,
# there is no external file downloader
def downloader(url:str,cfg:Config,io:IO) -> str:
    file_name = os.path.basename(urlparse(url).path)
    if cfg.download and io:
        if file_name:
            name,ext = os.path.splitext(file_name)

            downloaded_file_name = str(uuid.uuid4())[:8]+ext 
            fullpath = os.path.join(cfg.output_path,downloaded_file_name)
                
            try:
                io.write_line(status("Downloading",f"{file_name}"))
                request.urlretrieve(url,fullpath)
                io.write_line(status("Downloaded",f'successfully downloaded "{file_name}" -> "{downloaded_file_name}"'))
                return name,downloaded_file_name
            except Exception as e:
                io.write_line(error(f"cannot download {file_name}: {e}"))
        else:
            io.write_line(error(f"unvalid {url}"))
    else:
        return file_name,url



import os,sys
import time

from notion2md.console.formatter import *
from notion2md.config import set_config,get_config
from notion2md.notion_api import get_children
from notion2md.convertor.block import BlockConvertor

from cleo import option
from cleo import Command

class ExportBlockCommand(Command):
    name = "block"
    description = "Export a Notion block object to markdown."

    options = [
        option("url", "u", "The url of Notion block object.", flag=False),
        option("id", "i", "The id of Notion block object.", flag=False),
        option("name", "N", "The name of Notion block object", flag=False),
        option("path", "p", "The path to save exported markdown file.", flag=False),
        option("download", None, "Download files/images inside of the block object", flag=True),
    ]
    help = """The block command retrieves and exports Notion <highlight>block object</highlight> to markdownfile.

By default, the <highlight>id</highlight> of the block will be the name of markdown file, but if you path a value after <code>--name/-n</code>, the value will be the name of the markdown file.
    
It will save the exported markdown file in the path(<highlight>"notion-output/"</highlight> or <highlight>specific path you entered</highlight>)

By default, it won't save files/images in the block object, but if you pass <code>--download</code> option, it will download files/images in the path(<highlight>"notion-output/"</highlight> or <highlight>specific path you entered</highlight>)
    """
    def handle(self):
        set_config(**self.option())
        config = get_config()
        if not config.target_id:
            self.line_error(error("Notion2Md requires either id or url."))
            sys.exit(1)
        start_time = time.time()
        exporter = BlockConvertor(self.io,config)
        #Directory Checking and Creating
        if not os.path.exists(config.output_path):
            os.mkdir(config.output_path)
        #Get actual blocks
        self.line("")
        self.line(status("Retrieving",f"blocks from '{config.target_id}'"))
        blocks = get_children(config.target_id)
        #Write(Export) Markdown file
        with open(os.path.join(config.output_path,config.file_name + '.md'),'w',encoding="utf-8") as output:
            output.write(exporter.convert(blocks))
            
        #Result and Time Check
        self.line(status("Converted", f"{str(len(blocks))} blocks to markdown in {time.time() - start_time:.2f}s"))
        self.line(status("Exported", f'"{config.file_name}.md" in "{os.path.abspath(config.output_path)}/"'))
        self.line("")


from notion2md.convertor.block import BlockConvertor
from notion2md.config import set_config,get_config
from notion2md.notion_api import get_children

import os

def block_markdown_exporter(**kargs):
    set_config(**kargs)
    config = get_config()
    #Directory Checking and Creating
    if not os.path.exists(config.output_path):
        os.mkdir(config.output_path)
    #Get actual blocks
    blocks = get_children(config.target_id)
    #Write(Export) Markdown file
    with open(os.path.join(config.output_path,config.file_name + '.md'),'w',encoding="utf-8") as output:
        output.write(BlockConvertor(config).convert(blocks))
    
def block_string_exporter(**kargs):
    set_config(**kargs)
    config = get_config()
    blocks = get_children(config.target_id)
    return BlockConvertor(config).to_string(blocks)


from cleo.config import ApplicationConfig as BaseApplicationConfig
from clikit.api.formatter import Style
from clikit.api.args.format.argument import Argument
from clikit.api.args.format.option import Option
from clikit.api.event import PRE_HANDLE
from clikit.api.event import PRE_RESOLVE
from clikit.config.default_application_config import DefaultApplicationConfig
from clikit.handler.help import HelpTextHandler
from clikit.resolver.help_resolver import HelpResolver

class ApplicationConfig(BaseApplicationConfig):
    def configure(self):
        self.set_io_factory(self.create_io)
        self.add_event_listener(PRE_RESOLVE, self.resolve_help_command)
        self.add_event_listener(PRE_HANDLE, self.print_version)

        self.add_option("help", "h", Option.NO_VALUE, "Display this help message")
        self.add_option("quiet", "q", Option.NO_VALUE, "Do not output any message")
        self.add_option(
            "verbose",
            "v",
            Option.OPTIONAL_VALUE,
            "Increase the verbosity of messages: "
            '"-v" for normal output, '
            '"-vv" for more verbose output '
            'and "-vvv" for debug',
        )
        self.add_option(
            "version", "V", Option.NO_VALUE, "Display this application version"
        )
        self.add_option("ansi", None, Option.NO_VALUE, "Force ANSI output")
        self.add_option("no-ansi", None, Option.NO_VALUE, "Disable ANSI output")
        self.add_option(
            "no-interaction",
            "n",
            Option.NO_VALUE,
            "Do not ask any interactive question",
        )

        with self.command("help") as c:
            c.set_description("Display the manual of a command")
            c.add_argument(
                "command", Argument.OPTIONAL | Argument.MULTI_VALUED, "The command name"
            )
            c.set_handler(HelpTextHandler(HelpResolver()))

        self.add_style(Style("status").fg("green").bold())
        self.add_style(Style("error").fg("red").bold())
        self.add_style(Style("code").fg("green"))
        self.add_style(Style("highlight").fg("blue").bold())




from notion_client import Client #,AsyncClient
import os,sys
from notion2md.console.formatter import *

try:
    notion_client_object = Client(auth=os.environ["NOTION_TOKEN"])
except:
    print("Notion Integration Token is not found")
    print(
    """
        Welcome to notion2md!

        To get started, you need to save your Notion Integration Token.
        Find your token at

            https://www.notion.so/my-integrations

        Then run shell command:

            $export NOTION_TOKEN="<Your Token>"
        
        If you want to save this environment variable after reboot,
        put upper command in your shell resource(ex: .bashrc or .zshrc)
    """
    )
    sys.exit(1)
# notion_async_client = AsyncClient(auth=os.environ["NOTION_TOKEN"])

def get_children(parent_id):
    return notion_client_object.blocks.children.list(parent_id)['results']


from .richtext import richtext_convertor
from .file import downloader

from notion2md.notion_api import get_children
from notion2md.config import Config

import concurrent.futures


class BlockConvertor:
    def __init__(self,config,io=None):
        self._config = config
        self._io = io
    
    @property
    def config(self):
        return self._config
    
    @config.setter
    def config(self,**kargs):
        self._config = Config(**kargs)
    
    def convert(self,blocks:dict) -> str:
        outcome_blocks:str = ""
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(self.block_convertor,blocks)
            outcome_blocks = "".join([result for result in results])
        return outcome_blocks
    
    def block_convertor(self,block:dict,depth=0) -> str:
        outcome_block:str = ""
        block_type = block['type']
        #Special Case: Block is blank
        if block_type == "paragraph" and not block['has_children'] and not block[block_type]['text']:
            outcome_block = blank() +"\n\n"
        else:
            if block_type in BLOCK_TYPES:
                outcome_block = BLOCK_TYPES[block_type](self.information_collector(block[block_type])) + "\n\n"
            else:
                outcome_block = f"[{block_type} is not supported]\n\n"
            if block['has_children']:
                if block_type == "child_page":
                    #call make_child_function
                    pass
                elif block_type == 'table':
                    depth += 1
                    child_blocks = get_children(block['id'])
                    table_list = []
                    for cell_block in child_blocks:
                        cell_block_type = cell_block['type']
                        table_list.append(BLOCK_TYPES[cell_block_type](self.information_collector(cell_block[cell_block_type])))
                    # convert to markdown table
                    for index,value in enumerate(table_list):
                        if index == 0:
                            outcome_block = " | " + " | ".join(value) + " | " + "\n"
                            outcome_block += " | " + " | ".join(['----'] * len(value)) + " | " + "\n"
                            continue
                        outcome_block += " | " + " | ".join(value) + " | " + "\n"
                    outcome_block += "\n"
                else:
                    depth += 1
                    child_blocks = get_children(block['id'])
                    for block in child_blocks:
                        outcome_block += "\t"*depth + self.block_convertor(block,depth)
        return outcome_block

    def information_collector(self,payload:dict) -> dict:
        information = dict()
        if "text" in payload:
            information['text'] = richtext_convertor(payload['text'])
        if "icon" in payload:
            information['icon'] = payload['icon']['emoji']
        if "checked" in payload:
            information['checked'] = payload['checked']
        if "expression" in payload:
            information['text'] = payload['expression']
        if "url" in payload:
            information['url'] = payload['url']
        if "caption" in payload:
            information['caption'] = richtext_convertor(payload['caption'])
        if "external" in payload:
            information['url'] = payload['external']['url']
            name,file_path = downloader(information['url'],self._io,self._config)
            information['file_name'] = name
            information['file_path'] = file_path
        if "language" in payload:
            information['language'] = payload['language']
        # interal url
        if "file" in payload:
            information['url'] = payload['file']['url']
            name,file_path = downloader(information['url'],self._io,self._config)
            information['file_name'] = name
            information['file_path'] = file_path
        # table cells
        if "cells" in payload:
            information['cells'] = payload['cells']

        return information
    
    def to_string(self,blocks:dict) -> str:
        return self.convert(blocks)

# Converting Methods
def paragraph(information:dict) -> str:
    return information['text']

def heading_1(information:dict) -> str:
    return f"# {information['text']}"

def heading_2(information:dict) -> str:
    return f"## {information['text']}"

def heading_3(information:dict) -> str:
    return f"### {information['text']}"

def callout(information:dict) -> str:
    return f"{information['icon']} {information['text']}"

def quote(information:dict) -> str:
    return f"> {information['text']}"

#toggle item will be changed as bulleted list item
def bulleted_list_item(information:dict) -> str:
    return f"- {information['text']}"

# numbering is not supported
def numbered_list_item(information:dict) -> str:
    """
    input: item:dict = {"number":int, "text":str}
    """
    return f"1. {information['text']}"

def to_do(information:dict) -> str:
    """
    input: item:dict = {"checked":bool, "test":str}
    """
    return f"- {'[x]' if information['checked'] else '[ ]'} {information['text']}"

#not yet supported
#child_database will be changed as child page
# def child_page(information:dict) -> str:
#     """
#     input: item:dict = {"id":str,"text":str}
#     """
#     #make_page(information['id'])
#     text = information['text']
#     return f'[{text}]({text})'

def code(information:dict) -> str:
    """
    input: item:dict = {"language":str,"text":str}
    """
    return f"\n```{information['language']}\n{information['text']}\n```"

def embed(information:dict) -> str:
    """
    input: item:dict ={"url":str,"text":str}
    """
    return f"[{information['url']}]({information['url']})"

def image(information:dict) -> str:
    """
    input: item:dict ={"url":str,"text":str,"caption":str}
    """
    # name,file_path = downloader(information['url'])

    if information['caption']:
        return f"![{information['file_name']}]({information['file_path']})\n\n{information['caption']}"
    else:
        return f"![{information['file_name']}]({information['file_path']})"

def file(information:dict) -> str:
    # name,file_path = downloader(information['url'])
    return f"[{information['file_name']}]({information['file_path']})"

def bookmark(information:dict) -> str:
    """
    input: item:dict ={"url":str,"text":str,"caption":str}
    """
    if information['caption']:
        return f"![{information['url']}]({information['url']})\n\n{information['caption']}"
    else:
        return f"![{information['url']}]({information['url']})"

def equation(information:dict) -> str:
    return f"$$ {information['text']} $$"

def divider(information:dict) -> str:
    return f"---"

def blank() -> str:
    return "<br/>"

def table_row(information:list) -> list:
    """
    input: item:list = [[richtext],....]
    """
    column_list = []
    for column in information['cells']:
        column_list.append(richtext_convertor(column))
    return column_list

# Since Synced Block has only child blocks, not name, it will return blank
def synced_block(information:list) -> str:
    return "[//]: # (Synced Block)"

# Block type map
BLOCK_TYPES = {
    "paragraph": paragraph,
    "heading_1": heading_1,
    "heading_2": heading_2,
    "heading_3": heading_3,
    "callout": callout,
    "toggle":bulleted_list_item,
    "quote": quote,
    "bulleted_list_item": bulleted_list_item,
    "numbered_list_item": numbered_list_item,
    "to_do": to_do,
    # "child_page": child_page,
    "code": code,
    "embed": embed,
    "image": image,
    "bookmark": bookmark,
    "equation": equation,
    "divider": divider,
    "file": file,
    'table_row': table_row,
    'synced_block':synced_block
}


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['notion2md',
 'notion2md.console',
 'notion2md.console.commands',
 'notion2md.console.config',
 'notion2md.convertor',
 'notion2md.exporter']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=0.8.1,<0.9.0', 'notion-client>=0.7.1']

entry_points = \
{'console_scripts': ['notion2md = notion2md.console.application:main']}

setup_kwargs = {
    'name': 'notion2md',
    'version': '2.5.0',
    'description': 'Notion Markdown Exporter with Python Cli',
    'long_description': '![Notion2Md logo - an arrow pointing from "N" to "MD"](Notion2md.jpg)\n\n<br/>\n\n## About Notion2Md\n\n[![PyPI version](https://badge.fury.io/py/notion2md.svg)](https://badge.fury.io/py/notion2md)\n<a href="https://hits.seeyoufarm.com"><img src="https://hits.seeyoufarm.com/api/count/incr/badge.svg?url=https%3A%2F%2Fgithub.com%2Fecho724%2Fnotion2md&count_bg=%23949191&title_bg=%23555555&icon=github.svg&icon_color=%23E7E7E7&title=visited&edge_flat=false"/></a>\n\n- Notion Markdown Exporter using **official notion api** by [notion-sdk-py](https://github.com/ramnes/notion-sdk-py)\n\n## API Key(Token)\n\n- Before getting started, create [an integration and find the token](https://www.notion.so/my-integrations). → [Learn more about authorization](https://developers.notion.com/docs/authorization).\n\n- Then save your api key(token) as your os environment variable\n\n```Bash\n$ export NOTION_TOKEN="{your integration token key}"\n```\n\n## Install\n\n```Bash\n$ pip install notion2md\n```\n\n## Usage: Shell Command\n\n![Terminal output of the `notion2md -h` command](notion2md_options.png)\n\n- Notion2md requires either `id` or `url` of the Notion page/block\n- **Download** option will download files/images in the `path` directory\n\n```Bash\nnotion2md --download -n post -p ~/MyBlog/content/posts -u https://notion.so/...\n```\n\n- This command will generate "**post.md**" in your \'**~/MyBlog/content/posts**\' directory\n\n## Usage: Python\n\n```Python\nfrom notion2md.exporter import block_markdown_exporter, block_string_exporter\n\n# block_markdown_exporter will make markdown file on your output path\nblock_markdown_exporter(id=\'...\',path=\'...\',download=True)\n\n# block_string_exporter will return output as String type\nmd = block_string_exporter(id=\'...\',path=\'...\')\n```\n\n## To-do\n\n- [x] Download file object(image and files)\n- [x] Table blocks\n- [x] Synced Block\n- [ ] Page Exporter\n- [ ] Database Exporter\n- [ ] Child page\n- [ ] Column List and Column Blocks\n\n## Contribution\n\nPlease read [Contribution Guide](CONTRIBUTION.md)\n\n## Donation\n\nIf you think **Notion2Md** is helpful to you, you can support me here:\n\n<a href="https://www.buymeacoffee.com/echo724" target="_blank"><img src="https://cdn.buymeacoffee.com/buttons/v2/default-yellow.png" alt="Buy Me A Coffee" style="height: 54px;" height="54"></a>\n\n## License\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'echo724',
    'author_email': 'eunchan1001@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/echo724/notion2md.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



from notion2md import __version__

from cleo import Application as BaseApplication

from .commands.export_block import ExportBlockCommand

from .config import ApplicationConfig

class Application(BaseApplication):
    def __init__(self):
        super(Application,self).__init__(
            "notion2md",__version__,config=ApplicationConfig("notion2md",__version__)
        )

        block_command = ExportBlockCommand()

        self.add(block_command.default())
        

def main():
    Application().run()

if __name__ == "__main__":
    main()


#Link
def text_link(item:dict):
    """
    input: item:dict ={"content":str,"link":str}
    """
    return f"[{item['content']}]({item['link']['url']})"

#Annotations
def bold(content:str):
    return f"**{content}**"

def italic(content:str):
    return f"*{content}*"

def strikethrough(content:str):
    return f"~~{content}~~"

def underline(content:str):
    return f"<u>{content}</u>"

def code(content:str):
    return f"`{content}`"

def color(content:str,color):
    return f"<span style='color:{color}'>{content}</span>"

def equation(content:str):
    return f"$ {content} $"

annotation_map = {
    "bold": bold,
    "italic": italic,
    "strikethrough": strikethrough,
    "underline": underline,
    "code": code,
}

#Mentions
def _mention_link(content,url):
    return f"([{content}]({url}])"

def user(information:dict):
    return f"({information['content']})"

def page(information:dict):
    return _mention_link(information['content'], information['url'])

def date(information:dict):
    return f"({information['content']})"

def database(information:dict):
    return _mention_link(information['content'], information['url'])

def mention_information(payload:dict):
    information = dict()
    if payload['href']:
        information['url'] = payload['href']
        if payload['plain_text'] != "Untitled":
            information['content'] = payload['plain_text']
        else:
            information['content'] = payload['href']
    else:
        information['content'] = payload['plain_text']
    
    return information

mention_map = {
    "user": user,
    "page": page,
    "database": database,
    "date": date
}

def richtext_word_converter(richtext:dict) -> str:
    outcome_word = ""
    plain_text = richtext["plain_text"]
    if richtext['type'] == "equation":
        outcome_word = equation(plain_text)
    elif richtext['type'] == "mention":
        mention_type = richtext['mention']['type']
        if mention_type in mention_map:
            outcome_word = mention_map[mention_type](mention_information(richtext))
    else:
        if richtext["href"]:
            outcome_word = text_link(richtext["text"])
        else:
            outcome_word = plain_text
        annot = richtext["annotations"]
        for key,transfer in annotation_map.items():
            if richtext["annotations"][key]:
                outcome_word = transfer(outcome_word)
        if annot["color"] != "default":
            outcome_word = color(outcome_word,annot["color"])
    return outcome_word


def richtext_convertor(richtext_list:list) -> str:
    outcome_sentence = ""
    for richtext in richtext_list:
        outcome_sentence += richtext_word_converter(richtext)
    return outcome_sentence